"""Typing support for Joplin's data API."""

from dataclasses import dataclass, field, fields
from datetime import datetime
import enum
import mimetypes
from typing import (
    Generic,
    List,
    MutableMapping,
    Optional,
    Union,
    Set,
    TypeVar,
)
import uuid


# Datatypes used by the Joplin API. Needed for arbitrary kwargs.
JoplinTypes = Union[float, int, str]
# Kwargs mapping of the datatypes.
JoplinKwargs = MutableMapping[str, JoplinTypes]


class EventChangeType(enum.IntEnum):
    # https://joplinapp.org/api/references/rest_api/#properties-4
    CREATED = 1
    UPDATED = 2
    DELETED = 3


class ItemType(enum.IntEnum):
    # https://joplinapp.org/api/references/rest_api/#item-type-ids
    NOTE = 1
    FOLDER = 2
    SETTING = 3
    RESOURCE = 4
    TAG = 5
    NOTE_TAG = 6
    SEARCH = 7
    ALARM = 8
    MASTER_KEY = 9
    ITEM_CHANGE = 10
    NOTE_RESOURCE = 11
    RESOURCE_LOCAL_STATE = 12
    REVISION = 13
    MIGRATION = 14
    SMART_FILTER = 15
    COMMAND = 16


class MarkupLanguage(enum.IntEnum):
    # https://discourse.joplinapp.org/t/api-body-vs-body-html/11697/4
    MARKDOWN = 1
    HTML = 2


def is_id_valid(id_: str) -> bool:
    """Check whether a string is a valid id."""
    if len(id_) == 32:
        # client ID
        # https://joplinapp.org/api/references/rest_api/#creating-a-note-with-a-specific-id
        # https://stackoverflow.com/a/11592279/7410886
        try:
            int(id_, 16)
            return True
        except ValueError:
            return False
    if len(id_) == 22:
        # server ID
        # https://joplinapp.org/help/dev/spec/server_items/
        return True
    return False


@dataclass
class BaseData:
    type_: Optional[ItemType] = None

    def __post_init__(self) -> None:
        # detect if data is encrypted
        encryption_applied = getattr(self, "encryption_applied", False)
        if encryption_applied is not None and bool(int(encryption_applied)):
            raise NotImplementedError("Encryption is not supported")

        # Cast the basic joplin API datatypes to more convenient datatypes.
        for field_ in fields(self):
            value = getattr(self, field_.name)
            if value is None:
                continue
            if field_.name in (
                "id",
                "parent_id",
                "share_id",
                "conflict_original_id",
                "master_key_id",
                "item_id",
            ):
                # Exclude integer and empty string IDs.
                if value and isinstance(value, str) and not is_id_valid(value):
                    raise ValueError("Invalid ID:", value)
            elif (
                field_.name.endswith("_time")
                or field_.name.endswith("Time")
                or field_.name
                in (
                    "todo_due",
                    "todo_completed",
                )
            ):
                try:
                    value_int = int(value)
                    casted_value = (
                        None
                        if value_int == 0
                        else datetime.utcfromtimestamp(value_int / 1000.0)
                    )
                    setattr(self, field_.name, casted_value)
                except ValueError:
                    # TODO: This is not spec conform.
                    casted_value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
                    setattr(self, field_.name, casted_value)
            elif field_.name in (
                "is_conflict",
                "is_todo",
                "encryption_applied",
                "is_shared",
                "encryption_blob_encrypted",
            ):
                setattr(self, field_.name, bool(int(value)))
            elif field_.name == "latitude":
                setattr(self, field_.name, float(value))
                if not (-90 <= float(value) <= 90):
                    raise ValueError("Invalid latitude:", value)
            elif field_.name == "longitude":
                setattr(self, field_.name, float(value))
                if not (-180 <= float(value) <= 180):
                    raise ValueError("Invalid longitude:", value)
            elif field_.name == "markup_language":
                setattr(self, field_.name, MarkupLanguage(int(value)))
            # elif field_.name == "order":
            # elif field_.name == "crop_rect":
            # elif field_.name == "icon":
            # elif field_.name == "filename":  # "file_extension"
            elif field_.name in ("item_type", "type_"):
                setattr(self, field_.name, ItemType(int(value)))
            elif field_.name == "type":
                setattr(self, field_.name, EventChangeType(int(value)))

    def assigned_fields(self) -> Set[str]:
        # Exclude "type_" for convenience.
        return set(
            field_.name
            for field_ in fields(self)
            if getattr(self, field_.name) is not None and field_.name != "type_"
        )

    @classmethod
    def fields(cls) -> Set[str]:
        # Exclude "type_" for convenience.
        return set(field_.name for field_ in fields(cls) if field_.name != "type_")

    @staticmethod
    def default_fields() -> Set[str]:
        return {"id", "parent_id", "title"}

    def __str__(self) -> str:
        # show only fields with values
        not_none_fields = ", ".join(
            f"{field.name}={getattr(self, field.name)}"
            for field in fields(self)
            if getattr(self, field.name) is not None
        )
        return f"{type(self).__name__}({not_none_fields})"


@dataclass
class NoteData(BaseData):
    """https://joplinapp.org/api/references/rest_api/#notes"""

    id: Optional[str] = None
    parent_id: Optional[str] = None
    title: Optional[str] = None
    body: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    is_conflict: Optional[bool] = None
    latitude: Optional[float] = None
    longitude: Optional[float] = None
    altitude: Optional[float] = None
    author: Optional[str] = None
    source_url: Optional[str] = None
    is_todo: Optional[bool] = None
    todo_due: Optional[datetime] = None
    todo_completed: Optional[datetime] = None
    source: Optional[str] = None
    source_application: Optional[str] = None
    application_data: Optional[str] = None
    order: Optional[float] = None
    user_created_time: Optional[datetime] = None
    user_updated_time: Optional[datetime] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    markup_language: Optional[MarkupLanguage] = None
    is_shared: Optional[bool] = None
    share_id: Optional[str] = None
    conflict_original_id: Optional[str] = None
    master_key_id: Optional[str] = None
    user_data: Optional[str] = None
    deleted_time: Optional[datetime] = None
    body_html: Optional[str] = None
    base_url: Optional[str] = None
    image_data_url: Optional[str] = None
    crop_rect: Optional[str] = None

    def serialize(self) -> str:
        # title is needed always to prevent problems with body
        # f. e. when there is a newline at start
        lines = ["" if self.title is None else self.title, ""]
        if self.body is not None:
            lines.extend([self.body, ""])
        for field_ in fields(self):
            if field_.name == "id":
                # ID is always required
                if self.id is None:
                    self.id = uuid.uuid4().hex
                lines.append(f"{field_.name}: {self.id}")
            elif field_.name == "markup_language":
                # required to get an editable note
                if self.markup_language is None:
                    self.markup_language = MarkupLanguage.MARKDOWN
                lines.append(f"{field_.name}: {self.markup_language}")
            elif field_.name == "source_application":
                if self.source_application is None:
                    self.source_application = "joppy"
                lines.append(f"{field_.name}: {self.source_application}")
            elif field_.name in ("title", "body"):
                pass  # handled before
            elif field_.name == "type_":
                self.item_type = ItemType.NOTE
                lines.append(f"{field_.name}: {self.item_type}")
            elif field_.name == "updated_time":
                # required, even if empty
                value_raw = getattr(self, field_.name)
                value = "" if value_raw is None else value_raw
                lines.append(f"{field_.name}: {value}")
            else:
                value_raw = getattr(self, field_.name)
                if value_raw is not None:
                    lines.append(f"{field_.name}: {value_raw}")
        return "\n".join(lines)


@dataclass
class NotebookData(BaseData):
    """https://joplinapp.org/api/references/rest_api/#folders"""

    id: Optional[str] = None
    title: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    user_created_time: Optional[datetime] = None
    user_updated_time: Optional[datetime] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    parent_id: Optional[str] = None
    is_shared: Optional[bool] = None
    share_id: Optional[str] = None
    master_key_id: Optional[str] = None
    icon: Optional[str] = None
    user_data: Optional[str] = None
    deleted_time: Optional[datetime] = None

    def serialize(self) -> str:
        lines = []
        if self.title is not None:
            lines.extend([self.title, ""])
        for field_ in fields(self):
            if field_.name == "id":
                # ID is always required
                if self.id is None:
                    self.id = uuid.uuid4().hex
                lines.append(f"{field_.name}: {self.id}")
            elif field_.name == "title":
                pass  # handled before
            elif field_.name == "type_":
                self.item_type = ItemType.FOLDER
                lines.append(f"{field_.name}: {self.item_type}")
            elif field_.name == "updated_time":
                # required, even if empty
                value_raw = getattr(self, field_.name)
                value = "" if value_raw is None else value_raw
                lines.append(f"{field_.name}: {value}")
            else:
                value_raw = getattr(self, field_.name)
                if value_raw is not None:
                    lines.append(f"{field_.name}: {value_raw}")
        return "\n".join(lines)


@dataclass
class ResourceData(BaseData):
    """https://joplinapp.org/api/references/rest_api/#resources"""

    id: Optional[str] = None
    title: Optional[str] = None
    mime: Optional[str] = None
    filename: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    user_created_time: Optional[datetime] = None
    user_updated_time: Optional[datetime] = None
    file_extension: Optional[str] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    encryption_blob_encrypted: Optional[bool] = None
    size: Optional[int] = None
    is_shared: Optional[bool] = None
    share_id: Optional[str] = None
    master_key_id: Optional[str] = None
    user_data: Optional[str] = None
    blob_updated_time: Optional[datetime] = None
    ocr_text: Optional[str] = None
    ocr_details: Optional[str] = None
    ocr_status: Optional[int] = None
    ocr_error: Optional[str] = None

    @staticmethod
    def default_fields() -> Set[str]:
        return {"id", "title"}

    def serialize(self) -> str:
        lines = []
        if self.title is not None:
            lines.extend([self.title, ""])
        # TODO: file_extension, size
        for field_ in fields(self):
            if field_.name == "id":
                # ID is always required
                if self.id is None:
                    self.id = uuid.uuid4().hex
                lines.append(f"{field_.name}: {self.id}")
            elif field_.name == "mime":
                # mime is always required
                if self.mime is None:
                    mime_type, _ = mimetypes.guess_type(self.filename or "")
                    self.mime = (
                        mime_type
                        if mime_type is not None
                        else "application/octet-stream"
                    )
                lines.append(f"{field_.name}: {self.mime}")
            elif field_.name == "title":
                pass  # handled before
            elif field_.name == "type_":
                self.item_type = ItemType.RESOURCE
                lines.append(f"{field_.name}: {self.item_type}")
            elif field_.name == "updated_time":
                # required, even if empty
                value_raw = getattr(self, field_.name)
                value = "" if value_raw is None else value_raw
                lines.append(f"{field_.name}: {value}")
            else:
                value_raw = getattr(self, field_.name)
                if value_raw is not None:
                    lines.append(f"{field_.name}: {value_raw}")
        return "\n".join(lines)


@dataclass
class RevisionData(BaseData):
    """https://joplinapp.org/help/api/references/rest_api/#revisions"""

    id: Optional[str] = None
    parent_id: Optional[str] = None
    item_type: Optional[ItemType] = None
    item_id: Optional[str] = None
    item_updated_time: Optional[datetime] = None
    title_diff: Optional[str] = None
    body_diff: Optional[str] = None
    metadata_diff: Optional[str] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    updated_time: Optional[datetime] = None
    created_time: Optional[datetime] = None

    @staticmethod
    def default_fields() -> Set[str]:
        return {"id"}


@dataclass
class TagData(BaseData):
    """https://joplinapp.org/api/references/rest_api/#tags"""

    id: Optional[str] = None
    title: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    user_created_time: Optional[datetime] = None
    user_updated_time: Optional[datetime] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    is_shared: Optional[bool] = None
    parent_id: Optional[str] = None
    user_data: Optional[str] = None

    def serialize(self) -> str:
        lines = []
        if self.title is not None:
            lines.extend([self.title, ""])
        for field_ in fields(self):
            if field_.name == "id":
                # ID is always required
                if self.id is None:
                    self.id = uuid.uuid4().hex
                lines.append(f"{field_.name}: {self.id}")
            elif field_.name == "title":
                pass  # handled before
            elif field_.name == "type_":
                self.item_type = ItemType.TAG
                lines.append(f"{field_.name}: {self.item_type}")
            elif field_.name == "updated_time":
                # required, even if empty
                value_raw = getattr(self, field_.name)
                value = "" if value_raw is None else value_raw
                lines.append(f"{field_.name}: {value}")
            else:
                value_raw = getattr(self, field_.name)
                if value_raw is not None:
                    lines.append(f"{field_.name}: {value_raw}")
        return "\n".join(lines)


@dataclass
class NoteTagData(BaseData):
    """Links a tag to a note."""

    id: Optional[str] = None
    note_id: Optional[str] = None
    tag_id: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    user_created_time: Optional[datetime] = None
    user_updated_time: Optional[datetime] = None
    encryption_cipher_text: Optional[str] = None
    encryption_applied: Optional[bool] = None
    is_shared: Optional[bool] = None

    def serialize(self) -> str:
        lines = []
        for field_ in fields(self):
            if field_.name == "id":
                # ID is always required
                if self.id is None:
                    self.id = uuid.uuid4().hex
                lines.append(f"{field_.name}: {self.id}")
            elif field_.name == "type_":
                self.item_type = ItemType.NOTE_TAG
                lines.append(f"{field_.name}: {self.item_type}")
            elif field_.name == "updated_time":
                # required, even if empty
                value_raw = getattr(self, field_.name)
                value = "" if value_raw is None else value_raw
                lines.append(f"{field_.name}: {value}")
            else:
                value_raw = getattr(self, field_.name)
                if value_raw is not None:
                    lines.append(f"{field_.name}: {value_raw}")
        return "\n".join(lines)


@dataclass
class EventData(BaseData):
    """https://joplinapp.org/api/references/rest_api/#events"""

    id: Optional[int] = None
    item_type: Optional[ItemType] = None
    item_id: Optional[int] = None
    type: Optional[EventChangeType] = None
    created_time: Optional[datetime] = None
    # source: Optional[int] = None
    # before_change_item: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        # Cast the basic joplin API datatypes to more convenient datatypes.
        if self.id is not None:
            self.id = int(self.id)

    @staticmethod
    def default_fields() -> Set[str]:
        return {"id", "item_type", "item_id", "type", "created_time"}


class LockType(enum.IntEnum):
    NONE = 0
    SYNC = 1
    EXCLUSIVE = 2


class LockClientType(enum.IntEnum):
    DESKTOP = 1
    MOBILE = 2
    CLI = 3


@dataclass
class LockData(BaseData):
    """
    https://joplinapp.org/help/dev/spec/sync_lock#lock-files
    https://github.com/laurent22/joplin/blob/b617a846964ea49be2ffefd31439e911ad84ed8c/packages/server/src/routes/api/locks.ts
    """

    id: Optional[str] = None
    type: Optional[LockType] = None
    clientId: Optional[str] = None
    clientType: Optional[LockClientType] = None
    updatedTime: Optional[datetime] = None


@dataclass
class UserData(BaseData):
    """
    https://joplinapp.org/help/dev/spec/server_user_status/
    https://github.com/laurent22/joplin/blob/b617a846964ea49be2ffefd31439e911ad84ed8c/packages/server/src/models/UserModel.ts#L117
    """

    id: Optional[str] = None
    email: Optional[str] = None
    password: Optional[str] = None
    is_admin: Optional[bool] = None
    full_name: Optional[str] = None
    created_time: Optional[datetime] = None
    updated_time: Optional[datetime] = None
    email_confirmed: Optional[bool] = None
    must_set_password: Optional[bool] = None
    account_type: Optional[int] = None  # TODO: enum
    can_upload: Optional[bool] = None
    max_item_size: Optional[int] = None
    max_total_item_size: Optional[int] = None
    total_item_size: Optional[int] = None
    can_share_folder: Optional[bool] = None
    can_share_note: Optional[bool] = None
    can_receive_folder: Optional[bool] = None
    enabled: Optional[bool] = None
    disabled_time: Optional[datetime] = None


AnyData = Union[
    EventData, NoteData, NotebookData, NoteTagData, ResourceData, RevisionData, TagData
]


T = TypeVar(
    "T",
    EventData,
    NoteData,
    NotebookData,
    ResourceData,
    RevisionData,
    TagData,
    LockData,
    UserData,
    str,
)


@dataclass
class DataList(Generic[T]):
    has_more: bool
    cursor: Optional[int] = None
    items: List[T] = field(default_factory=list)

    def __post_init__(self) -> None:
        # Cast the basic joplin API datatypes to more convenient datatypes.
        self.has_more = bool(self.has_more)



"""Helper functions for the API."""

import base64
from typing import Callable, List

import joppy.data_types as dt


def encode_base64(filepath: str) -> str:
    """Encode an arbitrary file to base64."""
    with open(filepath, "rb") as img_file:
        return base64.b64encode(img_file.read()).decode("utf-8")


def _unpaginate(
    func: Callable[..., dt.DataList[dt.T]], **query: dt.JoplinTypes
) -> List[dt.T]:
    """Calls an Joplin endpoint until it's response doesn't contain more data."""
    response = func(**query)
    items = response.items
    page = 1  # pages are one based
    while response.has_more:
        page += 1
        query["page"] = page
        response = func(**query)
        items.extend(response.items)
    return items



"""Tests for the Joplin server python API."""

import string
import time
from typing import cast
import unittest

from joppy.server_api import deserialize, LockError, ServerApi
import joppy.data_types as dt
from . import common, setup_joplin


API = None
SERVER = None


def setUpModule():  # pylint: disable=invalid-name
    global API, SERVER
    SERVER = setup_joplin.JoplinServer()
    # login only once to prevent
    # "429 Client Error: Too Many Requests for url: http://localhost:22300/login"
    API = ServerApi()


def tearDownModule():  # pylint: disable=invalid-name
    if SERVER is not None:
        SERVER.stop()


class ServerBase(common.Base):
    def setUp(self):
        self.api = cast(ServerApi, API)
        with self.api.sync_lock():  # needed for clearing all data
            super().setUp()

        self.api._acquire_sync_lock()
        if self.api.current_sync_lock is None:
            raise LockError("Couldn't aqcuire sync lock")

    def tearDown(self):
        super().tearDown()
        self.api._delete_own_lock()


class Note(ServerBase):
    def test_add(self):
        """Add a note to an existing notebook."""
        parent_id = self.api.add_notebook()
        id_ = self.api.add_note(parent_id=parent_id)

        notes = self.api.get_notes().items
        self.assertEqual(len(notes), 1)
        self.assertEqual(notes[0].id, id_)
        self.assertEqual(notes[0].parent_id, parent_id)

    def test_delete(self):
        """Add and then delete a note."""
        parent_id = self.api.add_notebook()
        id_ = self.api.add_note(parent_id)
        notes = self.api.get_notes()
        self.assertEqual(len(notes.items), 1)

        self.api.delete_note(id_=id_)
        self.assertEqual(self.api.get_notes().items, [])

    def test_get_note(self):
        """Get a specific note."""
        parent_id = self.api.add_notebook()
        id_ = self.api.add_note(parent_id=parent_id)
        note = self.api.get_note(id_=id_)
        self.assertEqual(note.type_, dt.ItemType.NOTE)

    def test_get_notes(self):
        """Get all notes."""
        parent_id = self.api.add_notebook()
        self.api.add_note(parent_id=parent_id)
        notes = self.api.get_notes()
        self.assertEqual(len(notes.items), 1)
        self.assertFalse(notes.has_more)

    def test_get_all_notes(self):
        """Get all notes, unpaginated."""
        parent_id = self.api.add_notebook()
        count = 101  # pagination seems to start at 100 items
        for _ in range(count):
            self.api.add_note(parent_id)
        self.assertEqual(len(self.api.get_all_notes()), count)

    def test_move(self):
        """Move a note from one notebook to another."""
        original_title = "test title"
        original_body = "test body"

        notebook_1_id = self.api.add_notebook()
        notebook_2_id = self.api.add_notebook()
        id_ = self.api.add_note(
            parent_id=notebook_1_id, title=original_title, body=original_body
        )

        note = self.api.get_note(id_=id_)
        self.assertEqual(note.parent_id, notebook_1_id)

        self.api.modify_note(id_=id_, parent_id=notebook_2_id)
        note = self.api.get_note(id_=id_)
        self.assertEqual(note.parent_id, notebook_2_id)

        # Ensure the original properties aren't modified.
        self.assertEqual(note.body, original_body)
        self.assertEqual(note.title, original_title)


class Notebook(ServerBase):
    def test_add(self):
        """Add a notebook."""
        id_ = self.api.add_notebook()

        notebooks = self.api.get_notebooks().items
        self.assertEqual(len(notebooks), 1)
        self.assertEqual(notebooks[0].id, id_)

    def test_get_notebook(self):
        """Get a specific notebook."""
        id_ = self.api.add_notebook()
        notebook = self.api.get_notebook(id_=id_)
        self.assertEqual(notebook.type_, dt.ItemType.FOLDER)

    def test_get_notebooks(self):
        """Get all notebooks."""
        self.api.add_notebook()
        notebooks = self.api.get_notebooks()
        self.assertEqual(len(notebooks.items), 1)
        self.assertFalse(notebooks.has_more)

    def test_get_all_notebooks(self):
        """Get all notebooks, unpaginated."""
        count = 101
        for _ in range(count):
            self.api.add_notebook()
        self.assertEqual(len(self.api.get_all_notebooks()), count)

    def test_move(self):
        """Move a root notebok to another notebook. It's now a subnotebook."""
        first_id = self.api.add_notebook()
        second_id = self.api.add_notebook()
        self.assertEqual(self.api.get_notebook(id_=first_id).parent_id, None)
        self.assertEqual(self.api.get_notebook(id_=second_id).parent_id, None)

        self.api.modify_notebook(id_=first_id, parent_id=second_id)
        self.assertEqual(self.api.get_notebook(id_=first_id).parent_id, second_id)
        self.assertEqual(self.api.get_notebook(id_=second_id).parent_id, None)


class Ping(ServerBase):
    def test_ping(self):
        """Ping should return the test string."""
        ping = self.api.ping()
        self.assertEqual(
            ping.json(), {"status": "ok", "message": "Joplin Server is running"}
        )


class Resource(ServerBase):
    @common.with_resource
    def test_add(self, filename):
        """Add a resource."""
        id_ = self.api.add_resource(filename=filename)

        resources = self.api.get_resources().items
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0].id, id_)

    @common.with_resource
    def test_add_to_note(self, filename):
        """Add a resource to an existing note."""
        parent_id = self.api.add_notebook()
        for file_ in ("test/grant_authorization_button.png", filename):
            with self.subTest(file_=file_):
                note_id = self.api.add_note(parent_id)
                resource_id = self.api.add_resource(filename=file_)
                self.api.add_resource_to_note(resource_id=resource_id, note_id=note_id)

                # Get the resource
                resource = self.api.get_resource(id_=resource_id)

                # Verify the markdown is correct (prefix "!" for images).
                note = self.api.get_note(id_=note_id)
                # TODO: Use "assertIsNotNone()" when
                # https://github.com/python/mypy/issues/5528 is resolved.
                assert resource.mime is not None
                image_prefix = "!" if resource.mime.startswith("image/") else ""
                self.assertEqual(
                    f"\n{image_prefix}[{file_}](:/{resource_id})", note.body
                )

    @common.with_resource
    def test_delete(self, filename):
        """Add and then delete a resource."""
        id_ = self.api.add_resource(filename=filename)
        resources = self.api.get_resources()
        self.assertEqual(len(resources.items), 1)

        self.api.delete_resource(id_=id_)
        self.assertEqual(self.api.get_resources().items, [])

    @common.with_resource
    def test_get_resource(self, filename):
        """Get metadata about a specific resource."""
        id_ = self.api.add_resource(filename=filename)
        resource = self.api.get_resource(id_=id_)
        self.assertEqual(resource.type_, dt.ItemType.RESOURCE)

    @common.with_resource
    def test_get_resource_file(self, filename):
        """Get a specific resource in binary format."""
        for file_ in ("test/grant_authorization_button.png", filename):
            id_ = self.api.add_resource(filename=file_)
            resource = self.api.get_resource_file(id_=id_)
            with open(file_, "rb") as resource_file:
                self.assertEqual(resource, resource_file.read())

    @common.with_resource
    def test_get_resources(self, filename):
        """Get all resources."""
        self.api.add_resource(filename=filename)
        resources = self.api.get_resources()
        self.assertEqual(len(resources.items), 1)
        self.assertFalse(resources.has_more)

    @common.with_resource
    def test_get_all_resources(self, filename):
        """Get all resources, unpaginated."""
        count = 5
        for _ in range(count):
            self.api.add_resource(filename=filename)
        self.assertEqual(len(self.api.get_all_resources()), count)

    # @common.with_resource
    # def test_modify_title(self, filename):
    #     """Modify a resource title."""
    #     id_ = self.api.add_resource(filename=filename)

    #     new_title = self.get_random_string(exclude=string.whitespace)
    #     self.api.modify_resource(id_=id_, title=new_title)
    #     self.assertEqual(self.api.get_resource(id_=id_).title, new_title)

    @common.with_resource
    def test_check_property_title(self, filename):
        """Check the title of a resource."""
        # newline seems to be stripped from the title
        title = self.get_random_string(exclude=string.whitespace)
        id_ = self.api.add_resource(filename=filename, title=title)
        resource = self.api.get_resource(id_=id_)
        self.assertEqual(resource.title, title)


class Tag(ServerBase):
    def test_add_no_note(self):
        """Tags can be added even without notes."""
        id_ = self.api.add_tag()

        tags = self.api.get_tags().items
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].id, id_)

    def test_add_to_note(self):
        """Add a tag to an existing note."""
        parent_id = self.api.add_notebook()
        note_id = self.api.add_note(parent_id)
        tag_id = self.api.add_tag()
        self.api.add_tag_to_note(tag_id=tag_id, note_id=note_id)

        tags = self.api.get_all_tags()
        self.assertEqual(len(tags), 1)

    def test_add_with_parent(self):
        """Add a tag as child for an existing note."""
        notebook_id = self.api.add_notebook()
        parent_id = self.api.add_note(notebook_id)
        id_ = self.api.add_tag(parent_id=parent_id)

        tags = self.api.get_tags().items
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].id, id_)
        self.assertEqual(tags[0].parent_id, parent_id)

    def test_get_tag(self):
        """Get a specific tag."""
        id_ = self.api.add_tag()
        tag = self.api.get_tag(id_=id_)
        self.assertEqual(tag.type_, dt.ItemType.TAG)

    def test_get_tags(self):
        """Get all tags."""
        self.api.add_tag()
        tags = self.api.get_tags()
        self.assertEqual(len(tags.items), 1)
        self.assertFalse(tags.has_more)

    def test_get_all_tags(self):
        """Get all tags, unpaginated."""
        count = 101
        for _ in range(count):
            self.api.add_tag()
        self.assertEqual(len(self.api.get_all_tags()), count)


class Lock(ServerBase):
    def test_sync_lock_different_id(self):
        self.assertEqual(len(self.api._get_all_locks()), 1)

        # Acquiring a second sync lock with a different client ID should succeed.
        second_api = ServerApi()
        with second_api.sync_lock():
            self.assertEqual(len(self.api._get_all_locks()), 2)
        self.assertEqual(len(self.api._get_all_locks()), 1)

    def test_sync_lock_same_id(self):
        self.assertEqual(len(self.api._get_all_locks()), 1)

        # Acquiring a second sync lock with the same client ID should fail.
        with self.assertRaises(LockError):
            second_api = ServerApi()
            second_api.client_id = self.api.client_id
            with second_api.sync_lock():
                self.fail("This should never happen")

        self.assertEqual(len(self.api._get_all_locks()), 1)

    @unittest.skipIf(not common.SLOW_TESTS, "Waiting for the timeout takes some time.")
    def test_sync_lock_refresh(self):
        locks = self.api._get_all_locks()
        self.assertEqual(len(locks), 1)
        updated_time_before = locks[0].updatedTime

        # refresh interval is not needed -> same lock
        self.api.get_all_notebooks()
        locks = self.api._get_all_locks()
        self.assertEqual(len(locks), 1)
        self.assertEqual(locks[0].updatedTime, updated_time_before)

        # some arbitrary request to refresh the sync lock
        time.sleep(self.api.lock_auto_refresh_interval.total_seconds() + 1)
        self.api.get_all_notebooks()

        locks = self.api._get_all_locks()
        self.assertEqual(len(locks), 1)
        self.assertGreater(locks[0].updatedTime, updated_time_before)

    @unittest.skipIf(not common.SLOW_TESTS, "Waiting for the timeout takes some time.")
    def test_sync_lock_expired(self):
        locks = self.api._get_all_locks()
        self.assertEqual(len(locks), 1)

        time.sleep(self.api.lock_ttl.total_seconds() + 1)
        # lock should be expired after self.api.lock_ttl.total_seconds()
        with self.assertRaises(LockError):
            self.api.get_all_notebooks()


class User(ServerBase):
    def test_get_current_user(self):
        """The current user should be available and have read and write permissions."""
        current_user = self.api.get_current_user()
        self.assertIsNotNone(current_user)
        self.assertTrue(current_user.enabled)
        self.assertTrue(current_user.can_upload)


class Deserialize(ServerBase):
    def test_note_only_metadata(self):
        body = "id: 0e8d296dbef34588b0de060630ad2582\ntype_: 1"
        result = deserialize(body)

        # TODO: use assertIsInstance() when mypy understands it
        assert isinstance(result, dt.NoteData)
        self.assertEqual(result.id, "0e8d296dbef34588b0de060630ad2582")
        self.assertIsNone(result.title)
        self.assertIsNone(result.body)

    def test_note_title_metadata(self):
        body = "note2\n\nid: 0e8d296dbef34588b0de060630ad2582\ntype_: 1"
        result = deserialize(body)

        # TODO: use assertIsInstance() when mypy understands it
        assert isinstance(result, dt.NoteData)
        self.assertEqual(result.id, "0e8d296dbef34588b0de060630ad2582")
        self.assertEqual(result.title, "note2")
        self.assertIsNone(result.body)

    def test_note_title_body_metadata(self):
        body = (
            "note2\n\nbody\n\nwith some\n\nnewlines\n\n"
            "id: 0e8d296dbef34588b0de060630ad2582\ntype_: 1"
        )
        result = deserialize(body)

        # TODO: use assertIsInstance() when mypy understands it
        assert isinstance(result, dt.NoteData)
        self.assertEqual(result.id, "0e8d296dbef34588b0de060630ad2582")
        self.assertEqual(result.title, "note2")
        self.assertEqual(result.body, "body\n\nwith some\n\nnewlines")



from contextlib import contextmanager
import datetime
import logging
import pathlib
import time
from typing import Any, cast, Dict, List, Optional, Tuple, Union
import uuid

import requests

import joppy.data_types as dt
from joppy import tools


# Use a global session object for better performance.
# Define it globally to avoid "ResourceWarning".
SESSION = requests.Session()

# Don't spam the log. See: https://stackoverflow.com/a/11029841/7410886.
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

LOGGER = logging.getLogger("joppy")


def deserialize(body: str) -> Optional[dt.AnyData]:
    """Deserialize server data from string to a known data type."""

    # https://github.com/laurent22/joplin/blob/b617a846964ea49be2ffefd31439e911ad84ed8c/packages/lib/models/BaseItem.ts#L549-L596
    def extract_metadata(serialized_metadata: str) -> Dict[Any, Any]:
        metadata = {}
        for line in serialized_metadata.split("\n"):
            key, value = line.split(": ", 1)
            if not value:
                continue
            metadata[key] = value
        return metadata

    metadata_splitted = body.rsplit("\n\n", 1)
    if len(metadata_splitted) == 1:
        # metadata only
        title = None
        note_body = None
        metadata = extract_metadata(metadata_splitted[0])
    else:
        metadata = extract_metadata(metadata_splitted[1])
        title_splitted = metadata_splitted[0].split("\n\n", 1)
        if len(title_splitted) == 1:
            # title + metadata
            title = title_splitted[0]
            note_body = None
        else:
            # title + body + metadata
            title = title_splitted[0]
            note_body = title_splitted[1]

    if title is not None:
        metadata["title"] = title
    if note_body is not None:
        metadata["body"] = note_body

    item_type = dt.ItemType(int(metadata["type_"]))
    if item_type == dt.ItemType.NOTE:
        return dt.NoteData(**metadata)
    elif item_type == dt.ItemType.FOLDER:
        return dt.NotebookData(**metadata)
    elif item_type == dt.ItemType.RESOURCE:
        return dt.ResourceData(**metadata)
    elif item_type == dt.ItemType.TAG:
        return dt.TagData(**metadata)
    elif item_type == dt.ItemType.NOTE_TAG:
        return dt.NoteTagData(**metadata)
    elif item_type == dt.ItemType.REVISION:
        # ignore revisions for now
        pass
    else:
        print("TODO: ", dt.ItemType(int(metadata["type_"])))
    return None


def add_suffix(string: str, suffix: str = ".md") -> str:
    """Add a suffix."""
    return str(pathlib.Path(string).with_suffix(suffix))


def remove_suffix(string: str) -> str:
    """Remove the suffix."""
    return str(pathlib.Path(string).with_suffix(""))


##############################################################################
# Base wrapper that manages the requests to the client REST API.
##############################################################################


class LockError(Exception):
    pass


class ApiBase:
    """Contains the basic requests of the server REST API."""

    def __init__(
        self,
        user: str = "admin@localhost",
        password: str = "admin",
        url: str = "http://localhost:22300",
    ) -> None:
        self.user = user
        self.url = url
        self.client_id = uuid.uuid4().hex
        self.current_sync_lock: Optional[dt.LockData] = None
        # TODO: Where to get it?
        # https://github.com/laurent22/joplin/blob/b617a846964ea49be2ffefd31439e911ad84ed8c/packages/lib/services/synchronizer/LockHandler.ts#L145
        self.lock_ttl = datetime.timedelta(seconds=60 * 3)
        self.lock_auto_refresh_interval = datetime.timedelta(seconds=60)

        # cookie is saved in session and used for the next requests
        self.post("/login", data={"email": self.user, "password": password})

        # check for compatible sync version
        sync_version = self.get_sync_version()
        if sync_version is None:
            LOGGER.warning("Sync version not found. Creating a new file.")
            # https://joplinapp.org/help/dev/spec/sync#sync-target-properties
            now = int(time.time() * 1000)
            sync_target_info = {
                "version": 3,
                "e2ee": {"value": False, "updatedTime": now},
                "activeMasterKeyId": {"value": False, "updatedTime": now},
                "masterKeys": [],
                "ppk": {},
                "appMinVersion": "3.0.0",
            }
            self.put("/api/items/root:/info.json:/content", json=sync_target_info)
        else:
            if sync_version != 3:
                raise NotImplementedError(
                    f"Only server version 3 is supported. Found version {sync_version}."
                )

    def get_sync_version(self) -> Optional[int]:
        try:
            response = self.get("/api/items/root:/info.json:/content")
            server_info = response.json()
            return int(server_info["version"])
        except requests.exceptions.HTTPError as error_json:
            if error_json.response.status_code == 404:
                try:
                    response = self.get("/api/items/root:/.sync/version.txt:/content")
                    return int(response.text)
                except requests.exceptions.HTTPError as error_txt:
                    if error_txt.response.status_code == 404:
                        return None  # in this case, a new one should be created
                    else:
                        raise error_txt
            else:
                raise error_json

    def _request(
        self,
        method: str,
        path: str,
        query: Optional[dt.JoplinKwargs] = None,
        data: Any = None,
        files: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> requests.models.Response:
        # bypass the lock for info, lock and login requests
        if not (
            path == "/login"
            or path.startswith("/api/locks")
            or "info.json" in path
            or ".sync/version.txt" in path
        ):
            if self.current_sync_lock is None:
                raise LockError("No sync lock. Acquire a lock before issuing requests.")
            assert self.current_sync_lock.updatedTime is not None  # for mypy
            if not self._is_lock_active(self.current_sync_lock.updatedTime):
                raise LockError(
                    "Sync lock expired. Delete the lock before issuing requests."
                )
            elif (
                self.current_sync_lock.updatedTime + self.lock_auto_refresh_interval
                < datetime.datetime.utcnow()
            ):
                LOGGER.debug("Refreshing sync lock.")
                self.current_sync_lock = self._add_lock()

        LOGGER.debug(f"API: {method} request: path={path}, query={query}, data={data}")
        if query is None:
            query = {}
        query_str = "&".join([f"{key}={val}" for key, val in query.items()])

        try:
            response: requests.models.Response = getattr(SESSION, method)(
                f"{self.url}{path}?{query_str}",
                data=data,
                files=files,
                json=json,
                headers=headers,
            )
            LOGGER.debug(f"API: response {response.text}")
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            err.args = err.args + (response.text,)
            raise
        return response

    def delete(
        self, path: str, query: Optional[dt.JoplinKwargs] = None
    ) -> requests.models.Response:
        """Convenience method to issue a delete request."""
        return self._request("delete", path, query=query)

    def get(
        self, path: str, query: Optional[dt.JoplinKwargs] = None
    ) -> requests.models.Response:
        """Convenience method to issue a get request."""
        return self._request("get", path, query=query)

    def post(
        self,
        path: str,
        data: Optional[dt.JoplinKwargs] = None,
        files: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
    ) -> requests.models.Response:
        """Convenience method to issue a post request."""
        return self._request("post", path, data=data, files=files, json=json)

    def put(
        self,
        path: str,
        data: Optional[Union[str, bytes]] = None,
        json: Optional[Dict[str, Any]] = None,
    ) -> requests.models.Response:
        """Convenience method to issue a put request."""
        return self._request(
            "put",
            path,
            data=data,
            json=json,
            headers={"Content-Type": "application/octet-stream"},
        )

    ##############################################################################
    # Lock functionality
    ##############################################################################

    def _add_lock(self) -> dt.LockData:
        """Add or refresh a lock."""
        data = {
            "type": dt.LockType.SYNC,
            "clientId": self.client_id,
            "clientType": dt.LockClientType.DESKTOP,
        }
        response = self.post("/api/locks", json=data)
        return dt.LockData(**response.json())

    def _delete_lock(
        self, lock_type: dt.LockType, client_type: dt.LockClientType, client_id: str
    ) -> None:
        """
        Delete a lock.
        https://joplinapp.org/help/dev/spec/sync_lock#lock-files
        """
        self.delete(f"/api/locks/{lock_type}_{client_type}_{client_id}")

    def _get_locks(self, **query: dt.JoplinTypes) -> dt.DataList[dt.LockData]:
        """
        Get locks, paginated.
        To get all locks (unpaginated), use "_get_all_locks()".
        """
        response = self.get("/api/locks", query=query).json()
        response["items"] = [dt.LockData(**item) for item in response["items"]]
        return dt.DataList[dt.LockData](**response)

    def _get_all_locks(self) -> List[dt.LockData]:
        """Get all locks, unpaginated."""
        return tools._unpaginate(self._get_locks)

    def _is_lock_active(self, updated_time: datetime.datetime) -> bool:
        return updated_time + self.lock_ttl > datetime.datetime.utcnow()

    def _acquire_sync_lock(self, tries: int = 1) -> None:
        """
        Acquire a sync lock.
        https://joplinapp.org/help/dev/spec/sync_lock#acquiring-a-sync-lock
        """
        # TODO: check sync target version
        # https://joplinapp.org/help/dev/spec/sync_lock#sync-target-migration

        def is_locked(
            check_lock_types: Tuple[dt.LockType, ...] = (
                dt.LockType.SYNC,
                dt.LockType.EXCLUSIVE,
            ),
        ) -> bool:
            # https://github.com/laurent22/joplin/blob/b617a846964ea49be2ffefd31439e911ad84ed8c/packages/lib/services/synchronizer/LockHandler.ts#L72-L75
            for lock in self._get_all_locks():
                if lock.type not in check_lock_types:
                    continue
                assert lock.updatedTime is not None
                if self._is_lock_active(lock.updatedTime):
                    if lock.type == dt.LockType.EXCLUSIVE:
                        return True
                    elif (
                        lock.type == dt.LockType.SYNC
                        and lock.clientId == self.client_id
                    ):
                        return True
                    # If there is no exclusive lock and no lock with our ID,
                    # sync is allowed.
            return False

        for delay in range(tries):
            if not is_locked():
                self.current_sync_lock = self._add_lock()
                if is_locked(check_lock_types=(dt.LockType.EXCLUSIVE,)):
                    # avoid race conditions
                    self._delete_own_lock()
                else:
                    return
            time.sleep(delay)
            LOGGER.debug("sync target is still locked")

    def _delete_own_lock(self) -> None:
        self.current_sync_lock = None
        self._delete_lock(dt.LockType.SYNC, dt.LockClientType.DESKTOP, self.client_id)

    @contextmanager
    def sync_lock(self) -> Any:
        self._acquire_sync_lock()
        if self.current_sync_lock is None:
            raise LockError("Couldn't aqcuire sync lock")
        yield
        self._delete_own_lock()


##############################################################################
# Specific classes
##############################################################################


class Note(ApiBase):
    def add_note(self, parent_id: str, **data: Any) -> str:
        """Add a note."""
        # Parent ID is required. Else the notes are created at root.
        note_data = dt.NoteData(parent_id=parent_id, **data)
        request_data = note_data.serialize()
        assert note_data.id is not None
        # Access all files by full path for now:
        # https://joplinapp.org/help/dev/spec/server_file_url_format
        self.put(
            f"/api/items/root:/{add_suffix(note_data.id)}:/content", data=request_data
        )
        return note_data.id

    def delete_note(self, id_: str) -> None:
        """Delete a note."""
        self.delete(f"/api/items/root:/{add_suffix(id_)}:")

    def get_note(self, id_: str) -> dt.NoteData:
        response = self.get(f"/api/items/root:/{add_suffix(id_)}:/content")
        return cast(dt.NoteData, deserialize(response.text))

    def get_notes(self, **query: dt.JoplinTypes) -> dt.DataList[dt.NoteData]:
        response = self.get("/api/items/root:/:/children", query=query).json()
        # TODO: Is this the best practice?
        notes = []
        for item in response["items"]:
            if item["name"].endswith(".md"):
                item_complete = self.get_note(remove_suffix(item["name"]))
                if isinstance(item_complete, dt.NoteData):
                    notes.append(item_complete)
        return dt.DataList(response["has_more"], response["cursor"], notes)

    def modify_note(self, id_: str, **data: Any) -> None:
        """Modify a note."""
        # TODO: Without fetching the orginal note, this would be replacing,
        # not modifying.
        id_server = add_suffix(id_)
        note_data = self.get_note(id_server)
        for key, value in data.items():
            setattr(note_data, key, value)
        request_data = note_data.serialize()
        self.put(f"/api/items/root:/{id_server}:/content", data=request_data)


class Notebook(ApiBase):
    def add_notebook(self, **data: Any) -> str:
        """Add a notebook."""
        notebook_data = dt.NotebookData(**data)
        request_data = notebook_data.serialize()
        assert notebook_data.id is not None
        self.put(
            f"/api/items/root:/{add_suffix(notebook_data.id)}:/content",
            data=request_data,
        )
        return notebook_data.id

    def delete_notebook(self, id_: str) -> None:
        """Delete a notebook."""
        self.delete(f"/api/items/root:/{add_suffix(id_)}:")

    def get_notebook(self, id_: str) -> dt.NotebookData:
        response = self.get(f"/api/items/root:/{add_suffix(id_)}:/content")
        return cast(dt.NotebookData, deserialize(response.text))

    def get_notebooks(self, **query: dt.JoplinTypes) -> dt.DataList[dt.NotebookData]:
        response = self.get("/api/items/root:/:/children", query=query).json()
        # TODO: Is this the best practice?
        notebooks = []
        for item in response["items"]:
            if item["name"].endswith(".md"):
                item_complete = self.get_notebook(remove_suffix(item["name"]))
                if isinstance(item_complete, dt.NotebookData):
                    notebooks.append(item_complete)
        return dt.DataList(response["has_more"], response["cursor"], notebooks)

    def modify_notebook(self, id_: str, **data: Any) -> None:
        """Modify a notebook."""
        # TODO: Without fetching the orginal notebook, this would be replacing,
        # not modifying.
        id_server = add_suffix(id_)
        notebook_data = self.get_notebook(id_server)
        for key, value in data.items():
            setattr(notebook_data, key, value)
        request_data = notebook_data.serialize()
        self.put(f"/api/items/root:/{id_server}:/content", data=request_data)


class Ping(ApiBase):
    def ping(self) -> requests.models.Response:
        """Ping the API."""
        return self.get("/api/ping")


class Resource(ApiBase):
    def add_resource(self, filename: str, **data: Any) -> str:
        """Add a resource."""

        # add the corresponding md item with metadata
        title = str(data.pop("title", filename))
        resource_data = dt.ResourceData(title=title, filename=filename, **data)
        request_data = resource_data.serialize()
        assert resource_data.id is not None
        self.put(
            f"/api/items/root:/{add_suffix(resource_data.id)}:/content",
            data=request_data,
        )

        # add the resource itself
        self.put(
            f"/api/items/root:/.resource/{resource_data.id}:/content",
            data=pathlib.Path(filename).read_bytes(),
        )

        return resource_data.id

    def delete_resource(self, id_: str) -> None:
        """Delete a resource."""
        # metadata
        self.delete(f"/api/items/root:/{add_suffix(id_)}:")
        # resource itself
        self.delete(f"/api/items/root:/.resource/{id_}:")

    def get_resource(self, id_: str) -> dt.ResourceData:
        """Get metadata about the resource with the given ID."""
        response = self.get(f"/api/items/root:/{add_suffix(id_)}:/content")
        return cast(dt.ResourceData, deserialize(response.text))

    def get_resource_file(self, id_: str) -> bytes:
        """Get the resource with the given ID in binary format."""
        return self.get(f"/api/items/root:/.resource/{id_}:/content").content

    def get_resources(self, **query: dt.JoplinTypes) -> dt.DataList[dt.ResourceData]:
        """
        Get resources, paginated.
        To get all resources (unpaginated), use "get_all_resources()".
        """
        response = self.get("/api/items/root:/:/children", query=query).json()
        # TODO: Is this the best practice?
        resources = []
        for item in response["items"]:
            if item["name"].endswith(".md"):
                item_complete = self.get_resource(remove_suffix(item["name"]))
                if isinstance(item_complete, dt.ResourceData):
                    resources.append(item_complete)
        return dt.DataList(response["has_more"], response["cursor"], resources)

    def modify_resource(self, id_: str, **data: Any) -> None:
        """Modify a resource."""
        # TODO: split in metadata and content?
        raise NotImplementedError("'modify_resource()' is not yet implemented")


class Revision(ApiBase):
    def delete_revision(self, id_: str) -> None:
        """Delete a revision."""
        self.delete(f"/api/items/root:/{add_suffix(id_)}:")

    def get_revision(self, id_: str) -> dt.RevisionData:
        """Get the revision with the given ID."""
        response = self.get(f"/api/items/root:/{add_suffix(id_)}:/content")
        return cast(dt.RevisionData, deserialize(response.text))

    def get_revisions(self, **query: Any) -> dt.DataList[dt.RevisionData]:
        response = self.get("/api/items/root:/:/children", query=query).json()
        # TODO: Is this the best practice?
        revisions = []
        for item in response["items"]:
            if item["name"].endswith(".md"):
                item_complete = self.get_revision(remove_suffix(item["name"]))
                if isinstance(item_complete, dt.RevisionData):
                    revisions.append(item_complete)
        return dt.DataList(response["has_more"], response["cursor"], revisions)


class Tag(ApiBase):
    def add_tag(self, **data: Any) -> str:
        """Add a tag."""
        tag_data = dt.TagData(**data)
        request_data = tag_data.serialize()
        assert tag_data.id is not None
        self.put(
            f"/api/items/root:/{add_suffix(tag_data.id)}:/content", data=request_data
        )
        return tag_data.id

    def delete_tag(self, id_: str) -> None:
        """Delete a tag."""
        self.delete(f"/api/items/root:/{add_suffix(id_)}:")

    def get_tag(self, id_: str) -> dt.TagData:
        """Get the tag with the given ID."""
        response = self.get(f"/api/items/root:/{add_suffix(id_)}:/content")
        return cast(dt.TagData, deserialize(response.text))

    def get_tags(self, **query: dt.JoplinTypes) -> dt.DataList[dt.TagData]:
        """
        Get tags, paginated.
        To get all tags (unpaginated), use "get_all_tags()".
        """
        response = self.get("/api/items/root:/:/children", query=query).json()
        # TODO: Is this the best practice?
        tags = []
        for item in response["items"]:
            if item["name"].endswith(".md"):
                item_complete = self.get_tag(remove_suffix(item["name"]))
                if isinstance(item_complete, dt.TagData):
                    tags.append(item_complete)
        return dt.DataList(response["has_more"], response["cursor"], tags)

    def modify_tag(self, id_: str, **data: Any) -> None:
        """Modify a tag."""
        # TODO: Without fetching the orginal tag, this would be replacing,
        # not modifying.
        id_server = add_suffix(id_)
        tag_data = self.get_tag(id_server)
        for key, value in data.items():
            setattr(tag_data, key, value)
        request_data = tag_data.serialize()
        self.put(f"/api/items/root:/{id_server}:/content", data=request_data)


class User(ApiBase):
    def get_users(self, **query: dt.JoplinTypes) -> dt.DataList[dt.UserData]:
        """
        Get users, paginated.
        To get all users (unpaginated), use "get_all_users()".
        """
        response = self.get("/api/users", query=query).json()
        response["items"] = [dt.UserData(**item) for item in response["items"]]
        return dt.DataList[dt.UserData](**response)


class ServerApi(Note, Notebook, Ping, Resource, Revision, Tag, User):
    """
    Collects all basic API functions and contains a few more useful methods.
    This should be the only class accessed from the users.
    """

    def add_tag_to_note(self, tag_id: str, note_id: str) -> str:
        """Add a tag to a given note."""
        note_tag_data = dt.NoteTagData(tag_id=tag_id, note_id=note_id)
        request_data = note_tag_data.serialize()
        assert note_tag_data.id is not None
        self.put(
            f"/api/items/root:/{add_suffix(note_tag_data.id)}:/content",
            data=request_data,
        )
        return note_tag_data.id

    def add_resource_to_note(self, resource_id: str, note_id: str) -> None:
        """Add a resource to a given note."""
        note = self.get_note(id_=note_id)
        resource = self.get_resource(id_=resource_id)
        # TODO: Use "assertIsNotNone()" when
        # https://github.com/python/mypy/issues/5528 is resolved.
        assert resource.mime is not None
        image_prefix = "!" if resource.mime.startswith("image/") else ""
        original_body = "" if note.body is None else note.body
        body_with_attachment = (
            f"{original_body}\n{image_prefix}[{resource.title}](:/{resource_id})"
        )
        self.modify_note(note_id, body=body_with_attachment)

    def delete_all_notes(self) -> None:
        """Delete all notes permanently."""
        for note in self.get_all_notes():
            assert note.id is not None
            self.delete_note(note.id)

    def delete_all_notebooks(self) -> None:
        """Delete all notebooks permanently."""
        for notebook in self.get_all_notebooks():
            assert notebook.id is not None
            self.delete_notebook(notebook.id)

    def delete_all_resources(self) -> None:
        """Delete all resources."""
        for resource in self.get_all_resources():
            assert resource.id is not None
            self.delete_resource(resource.id)

    def delete_all_revisions(self) -> None:
        """Delete all revisions."""
        for revision in self.get_all_revisions():
            assert revision.id is not None
            self.delete_revision(revision.id)

    def delete_all_tags(self) -> None:
        """Delete all tags."""
        for tag in self.get_all_tags():
            assert tag.id is not None
            self.delete_tag(tag.id)

    def get_all_notes(self) -> List[dt.NoteData]:
        """Get all notes, unpaginated."""
        return tools._unpaginate(self.get_notes)

    def get_all_notebooks(self) -> List[dt.NotebookData]:
        """Get all notebooks, unpaginated."""
        return tools._unpaginate(self.get_notebooks)

    def get_all_resources(self) -> List[dt.ResourceData]:
        """Get all resources, unpaginated."""
        return tools._unpaginate(self.get_resources)

    def get_all_revisions(self) -> List[dt.RevisionData]:
        """Get all revisions, unpaginated."""
        return tools._unpaginate(self.get_revisions)

    def get_all_tags(self) -> List[dt.TagData]:
        """Get all tags, unpaginated."""
        return tools._unpaginate(self.get_tags)

    def get_all_users(self) -> List[dt.UserData]:
        """Get all users, unpaginated."""
        return tools._unpaginate(self.get_users)

    def get_current_user(self) -> Optional[dt.UserData]:
        """https://joplinapp.org/help/dev/spec/server_user_status/#user-status"""
        current_user = None
        for user in self.get_all_users():
            if user.email == self.user:
                current_user = user
                break
        return current_user



"""Interface for the joplin data API."""

import copy
import json
import logging
import time
from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
    Union,
)
import urllib.parse

import requests

import joppy.data_types as dt
from joppy import tools


# Use a global session object for better performance.
# Define it globally to avoid "ResourceWarning".
SESSION = requests.Session()


# Don't spam the log. See: https://stackoverflow.com/a/11029841/7410886.
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

LOGGER = logging.getLogger("joppy")


##############################################################################
# Base wrapper that manages the requests to the client REST API.
##############################################################################


class ApiBase:
    """Contains the basic requests of the client REST API."""

    def __init__(self, token: str, url: str = "http://localhost:41184") -> None:
        self.url = url
        self.token = token

    def _request(
        self,
        method: str,
        path: str,
        query: Optional[dt.JoplinKwargs] = None,
        data: Optional[dt.JoplinKwargs] = None,
        files: Optional[Dict[str, Any]] = None,
    ) -> requests.models.Response:
        LOGGER.debug(
            f"API: {method} request: path={path}, query={query}, data={data}, "
            f"files={files}"
        )
        if data is not None and "id_" in data:
            # "id" is a reserved keyword in python, so don't use it.
            data["id"] = data.pop("id_")
        if query is None:
            query = {}
        query["token"] = self.token  # TODO: extending the dict may have side effects
        query_str = "&".join([f"{key}={val}" for key, val in query.items()])

        try:
            response: requests.models.Response = getattr(SESSION, method)(
                f"{self.url}{path}?{query_str}",
                json=data,
                files=files,
            )
            LOGGER.debug(f"API: response {response.text}")
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            err.args = err.args + (response.text,)
            raise
        return response

    def delete(
        self, path: str, query: Optional[dt.JoplinKwargs] = None
    ) -> requests.models.Response:
        """Convenience method to issue a delete request."""
        return self._request("delete", path, query=query)

    def get(
        self, path: str, query: Optional[dt.JoplinKwargs] = None
    ) -> requests.models.Response:
        """Convenience method to issue a get request."""
        return self._request("get", path, query=query)

    def post(
        self,
        path: str,
        data: Optional[dt.JoplinKwargs] = None,
        files: Optional[Dict[str, Any]] = None,
    ) -> requests.models.Response:
        """Convenience method to issue a post request."""
        return self._request("post", path, data=data, files=files)

    def put(
        self, path: str, data: Optional[dt.JoplinKwargs] = None
    ) -> requests.models.Response:
        """Convenience method to issue a put request."""
        return self._request("put", path, data=data)


##############################################################################
# The following classes contain all direct calls to a single endpoint.
# For further information, see: https://joplinapp.org/api/references/rest_api/
##############################################################################


class Event(ApiBase):
    """
    Events are supported since Joplin 2.4.5.
    See: https://github.com/laurent22/joplin/releases/tag/v2.4.5
    """

    def get_event(self, id_: int, **query: dt.JoplinTypes) -> dt.EventData:
        """Get the event with the given ID."""
        response = dt.EventData(**self.get(f"/events/{id_}", query=query).json())
        return response

    def get_events(self, **query: dt.JoplinTypes) -> dt.DataList[dt.EventData]:
        """
        Get events, paginated. To get all events (unpaginated), use
        "get_all_events()".
        """
        response = self.get("/events", query=query).json()
        response["items"] = [dt.EventData(**item) for item in response["items"]]
        return dt.DataList[dt.EventData](**response)


class Note(ApiBase):
    def add_note(self, **data: dt.JoplinTypes) -> str:
        """Add a note."""
        return str(self.post("/notes", data=data).json()["id"])

    def delete_note(self, id_: str, **query: dt.JoplinTypes) -> None:
        """Delete a note."""
        self.delete(f"/notes/{id_}", query=query)

    def get_note(self, id_: str, **query: dt.JoplinTypes) -> dt.NoteData:
        """Get the note with the given ID."""
        response = dt.NoteData(**self.get(f"/notes/{id_}", query=query).json())
        return response

    def get_notes(
        self,
        notebook_id: Optional[str] = None,
        resource_id: Optional[str] = None,
        tag_id: Optional[str] = None,
        **query: dt.JoplinTypes,
    ) -> dt.DataList[dt.NoteData]:
        """
        Get notes, paginated. If a notebook, resource or tag ID is given,
        return the corresponding notes. To get all notes (unpaginated), use
        "get_all_notes()".
        """
        if [notebook_id, resource_id, tag_id].count(None) < 2:
            raise ValueError("Too many IDs. Specify at most one.")
        notebook = "" if notebook_id is None else f"/folders/{notebook_id}"
        resource = "" if resource_id is None else f"/resources/{resource_id}"
        tag = "" if tag_id is None else f"/tags/{tag_id}"
        response = self.get(f"{notebook}{resource}{tag}/notes", query=query).json()
        response["items"] = [dt.NoteData(**item) for item in response["items"]]
        return dt.DataList[dt.NoteData](**response)

    def modify_note(self, id_: str, **data: dt.JoplinTypes) -> None:
        """Modify a note."""
        self.put(f"/notes/{id_}", data=data)


class Notebook(ApiBase):
    def add_notebook(self, **data: dt.JoplinTypes) -> str:
        """Add a notebook."""
        return str(self.post("/folders", data=data).json()["id"])

    def delete_notebook(self, id_: str, **query: dt.JoplinTypes) -> None:
        """Delete a notebook."""
        self.delete(f"/folders/{id_}", query=query)

    def get_notebook(self, id_: str, **query: dt.JoplinTypes) -> dt.NotebookData:
        """Get the notebook with the given ID."""
        response = dt.NotebookData(**self.get(f"/folders/{id_}", query=query).json())
        return response

    def get_notebooks(self, **query: dt.JoplinTypes) -> dt.DataList[dt.NotebookData]:
        """
        Get notebooks, paginated. To get all notebooks (unpaginated), use
        "get_all_notebooks()".
        """
        response = self.get("/folders", query=query).json()
        response["items"] = [dt.NotebookData(**item) for item in response["items"]]
        return dt.DataList[dt.NotebookData](**response)

    def modify_notebook(self, id_: str, **data: dt.JoplinTypes) -> None:
        """Modify a notebook."""
        self.put(f"/folders/{id_}", data=data)


class Ping(ApiBase):
    def ping(self) -> requests.models.Response:
        """Ping the API."""
        return self.get("/ping")


class Resource(ApiBase):
    def add_resource(self, filename: str, **data: dt.JoplinTypes) -> str:
        """Add a resource."""
        # Preserve the filename if there is no title specified.
        if data.get("title") is None:
            data["title"] = filename
        with open(filename, "rb") as infile:
            files = {
                "data": (json.dumps(filename), infile),
                "props": (None, json.dumps(data)),
            }
            return str(self.post("/resources", files=files).json()["id"])

    def delete_resource(self, id_: str) -> None:
        """Delete a resource."""
        self.delete(f"/resources/{id_}")

    def get_resource(self, id_: str, **query: dt.JoplinTypes) -> dt.ResourceData:
        """Get metadata about the resource with the given ID."""
        response = dt.ResourceData(**self.get(f"/resources/{id_}", query=query).json())
        return response

    def get_resource_file(self, id_: str) -> bytes:
        """Get the resource with the given ID in binary format."""
        return self.get(f"/resources/{id_}/file").content

    def get_resources(
        self, note_id: Optional[str] = None, **query: dt.JoplinTypes
    ) -> dt.DataList[dt.ResourceData]:
        """
        Get resources, paginated. If a note ID is given, return the corresponding
        resources. To get all resources (unpaginated), use "get_all_resources()".
        """
        note = "" if note_id is None else f"/notes/{note_id}"
        response = self.get(f"{note}/resources", query=query).json()
        response["items"] = [dt.ResourceData(**item) for item in response["items"]]
        return dt.DataList[dt.ResourceData](**response)

    def modify_resource(self, id_: str, **data: dt.JoplinTypes) -> None:
        """Modify a resource."""
        self.put(f"/resources/{id_}", data=data)


class Revision(ApiBase):
    """
    Revisions are supported since Joplin 2.13.2.
    See: https://github.com/laurent22/joplin/releases/tag/v2.13.2
    """

    def add_revision(
        self, item_id: str, item_type: dt.ItemType, **data: dt.JoplinTypes
    ) -> str:
        """Add a revision to an item."""
        now_unix_seconds = int(time.time())
        return str(
            self.post(
                "/revisions",
                data={
                    # There seem to be some undocumented required fields.
                    "item_id": item_id,
                    "item_type": item_type.value,
                    "item_updated_time": now_unix_seconds,
                    "item_created_time": now_unix_seconds,
                    **data,
                },
            ).json()["id"]
        )

    def delete_revision(self, id_: str) -> None:
        """Delete a revision."""
        self.delete(f"/revisions/{id_}")

    def get_revision(self, id_: str, **query: dt.JoplinTypes) -> dt.RevisionData:
        """Get the revision with the given ID."""
        response = dt.RevisionData(**self.get(f"/revisions/{id_}", query=query).json())
        return response

    def get_revisions(self, **query: dt.JoplinTypes) -> dt.DataList[dt.RevisionData]:
        """
        Get revisions, paginated. To get all revisions (unpaginated), use
        "get_all_revisions()".
        """
        response = self.get("/revisions", query=query).json()
        response["items"] = [dt.RevisionData(**item) for item in response["items"]]
        return dt.DataList[dt.RevisionData](**response)

    def modify_revision(self, id_: str, **data: dt.JoplinTypes) -> None:
        """Modify a revision."""
        self.put(f"/revisions/{id_}", data=data)


class Search(ApiBase):
    def search(
        self, **query: dt.JoplinTypes
    ) -> Union[
        dt.DataList[dt.NoteData],
        dt.DataList[dt.NotebookData],
        dt.DataList[dt.ResourceData],
        dt.DataList[dt.TagData],
        dt.DataList[str],
    ]:
        """Issue a search."""
        # Copy the dict, because its content gets changed.
        outer_query = copy.deepcopy(query)
        # The outer query is the query of the URL. The inner query is the query of the
        # Joplin search endpoint. To avoid special characters in the inner query that
        # would be recognized as outer query, encode it.
        outer_query["query"] = urllib.parse.quote(cast(str, outer_query["query"]))
        response = self.get("/search", query=outer_query).json()
        # TODO: add missing types
        items = response["items"]
        if "type" not in query or query["type"] == dt.ItemType.NOTE.name.lower():
            response["items"] = [dt.NoteData(**item) for item in items]
            return dt.DataList[dt.NoteData](**response)
        elif query["type"] == dt.ItemType.FOLDER.name.lower():
            response["items"] = [dt.NotebookData(**item) for item in items]
            return dt.DataList[dt.NotebookData](**response)
        elif query["type"] == dt.ItemType.RESOURCE.name.lower():
            response["items"] = [dt.ResourceData(**item) for item in items]
            return dt.DataList[dt.ResourceData](**response)
        elif query["type"] == dt.ItemType.TAG.name.lower():
            response["items"] = [dt.TagData(**item) for item in items]
            return dt.DataList[dt.TagData](**response)
        elif query["type"] == dt.ItemType.MASTER_KEY.name.lower():
            return dt.DataList[str](**response)
        raise NotImplementedError(f"Type {query['type']} not implemented, yet.")


class Tag(ApiBase):
    def add_tag(self, tag_id: Optional[str] = None, **data: dt.JoplinTypes) -> str:
        """
        Add a tag. If a tag is given, add the tag to a note.
        The data has to contain the note ID.
        """
        note = "" if tag_id is None else f"/{tag_id}/notes"
        # Don't put the response into a Tag object, since it contains
        # "note_id" and "tag_id", which are undocumented.
        return str(self.post(f"/tags{note}", data=data).json()["id"])

    def delete_tag(self, id_: str, note_id: Optional[str] = None) -> None:
        """Delete a tag. If a note is given, remove the tag from this note."""
        note = "" if note_id is None else f"/notes/{note_id}"
        self.delete(f"/tags/{id_}{note}")

    def get_tag(self, id_: str, **query: dt.JoplinTypes) -> dt.TagData:
        """Get the tag with the given ID."""
        response = dt.TagData(**self.get(f"/tags/{id_}", query=query).json())
        return response

    def get_tags(
        self, note_id: Optional[str] = None, **query: dt.JoplinTypes
    ) -> dt.DataList[dt.TagData]:
        """
        Get tags, paginated. If a note is given, return the corresponding tags.
        To get all tags (unpaginated), use "get_all_tags()".
        """
        note = "" if note_id is None else f"/notes/{note_id}"
        response = self.get(f"{note}/tags", query=query).json()
        response["items"] = [dt.TagData(**item) for item in response["items"]]
        return dt.DataList[dt.TagData](**response)

    def modify_tag(self, id_: str, **data: dt.JoplinTypes) -> None:
        """Modify a tag."""
        self.put(f"/tags/{id_}", data=data)


class ClientApi(Event, Note, Notebook, Ping, Resource, Revision, Search, Tag):
    """
    Collects all basic API functions and contains a few more useful methods.
    This should be the only class accessed from the users.
    """

    def add_tag_to_note(self, tag_id: str, note_id: str) -> None:
        """Add a tag to a given note."""
        note = self.get_note(id_=note_id, fields="id")
        assert note.id is not None
        self.add_tag(tag_id=tag_id, id_=note.id)

    def add_resource_to_note(self, resource_id: str, note_id: str) -> None:
        """Add a resource to a given note."""
        note = self.get_note(id_=note_id, fields="body")
        resource = self.get_resource(id_=resource_id, fields="title,mime")
        # TODO: Use "assertIsNotNone()" when
        # https://github.com/python/mypy/issues/5528 is resolved.
        assert resource.mime is not None
        image_prefix = "!" if resource.mime.startswith("image/") else ""
        body_with_attachment = (
            f"{note.body}\n{image_prefix}[{resource.title}](:/{resource_id})"
        )
        self.modify_note(note_id, body=body_with_attachment)

    def delete_all_notes(self) -> None:
        """Delete all notes permanently."""
        for note in self.get_all_notes():
            assert note.id is not None
            self.delete_note(note.id, permanent=1)

    def delete_all_notebooks(self) -> None:
        """Delete all notebooks permanently."""
        for notebook in self.get_all_notebooks():
            # Deleting the root notebooks is sufficient.
            if not notebook.parent_id:
                assert notebook.id is not None
                self.delete_notebook(notebook.id, permanent=1)

    def delete_all_resources(self) -> None:
        """Delete all resources."""
        for resource in self.get_all_resources():
            assert resource.id is not None
            self.delete_resource(resource.id)

    def delete_all_revisions(self) -> None:
        """Delete all revisions."""
        for revision in self.get_all_revisions():
            assert revision.id is not None
            self.delete_revision(revision.id)

    def delete_all_tags(self) -> None:
        """Delete all tags."""
        for tag in self.get_all_tags():
            assert tag.id is not None
            self.delete_tag(tag.id)

    def get_all_events(self, **query: dt.JoplinTypes) -> List[dt.EventData]:
        """Get all events, unpaginated."""
        return tools._unpaginate(self.get_events, **query)

    def get_all_notes(self, **query: dt.JoplinTypes) -> List[dt.NoteData]:
        """Get all notes, unpaginated."""
        return tools._unpaginate(self.get_notes, **query)

    def get_all_notebooks(self, **query: dt.JoplinTypes) -> List[dt.NotebookData]:
        """Get all notebooks, unpaginated."""
        return tools._unpaginate(self.get_notebooks, **query)

    def get_all_resources(self, **query: dt.JoplinTypes) -> List[dt.ResourceData]:
        """Get all resources, unpaginated."""
        return tools._unpaginate(self.get_resources, **query)

    def get_all_revisions(self, **query: dt.JoplinTypes) -> List[dt.RevisionData]:
        """Get all revisions, unpaginated."""
        return tools._unpaginate(self.get_revisions, **query)

    def get_all_tags(self, **query: dt.JoplinTypes) -> List[dt.TagData]:
        """Get all tags, unpaginated."""
        return tools._unpaginate(self.get_tags, **query)

    def search_all(
        self, **query: dt.JoplinTypes
    ) -> List[Union[dt.NoteData, dt.NotebookData, dt.ResourceData, dt.TagData]]:
        """Issue a search and get all results, unpaginated."""
        return tools._unpaginate(self.search, **query)  # type: ignore



import pathlib
from setuptools import setup


setup(
    name="joppy",
    version="1.0.0",
    packages=["joppy"],
    # https://stackoverflow.com/a/70386281/7410886
    package_data={
        "joppy": ["py.typed"],
    },
    description="Python API for Joplin",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/marph91/joppy",
    author="Martin Dörfelt",
    author_email="martin.d@andix.de",
    license="Mozilla Public License version 2.0",
    # https://pypi.org/classifiers
    classifiers=[
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Typing :: Typed",
    ],
    python_requires=">=3.7",
    install_requires=["requests"],
)



"""Tests for the Joplin client python API."""

from datetime import datetime
import mimetypes
import os
import random
import re
import string
from typing import Any, Mapping
import unittest

import requests
import urllib3

from joppy import tools
from joppy.client_api import ClientApi
import joppy.data_types as dt
from . import common, setup_joplin


PROFILE = os.path.join(os.getcwd(), "test_profile")
API_TOKEN = ""  # Don't use the API token from env to avoid data loss.
APP = None


def setUpModule():  # pylint: disable=invalid-name
    # TODO: When splitting in multiple files, this needs to be run at start of the
    # testsuite.
    global API_TOKEN, APP
    if not API_TOKEN:
        app_path = "./joplin.AppImage"
        setup_joplin.download_joplin_client(app_path)
        APP = setup_joplin.JoplinClient(app_path, PROFILE)
        API_TOKEN = APP.api_token


def tearDownModule():  # pylint: disable=invalid-name
    if APP is not None:
        APP.stop()


class ClientBase(common.Base):
    def setUp(self):
        self.api = ClientApi(token=API_TOKEN)
        super().setUp()


class Event(ClientBase):
    def generate_event(self) -> None:
        """Generate an event and wait until it's available."""
        events = self.api.get_events()
        assert events.cursor is not None
        events_before = len(self.api.get_all_events(cursor=events.cursor))
        # For now only notes trigger events:
        # https://joplinapp.org/api/references/rest_api/#events
        self.api.add_notebook()
        self.api.add_note()

        def compare_event_count():
            assert events.cursor is not None
            event_count = len(self.api.get_all_events(cursor=events.cursor))
            return event_count if event_count == events_before + 1 else None

        # Wait until the event is available.
        setup_joplin.wait_for(compare_event_count, interval=0.1, timeout=1)

    def test_get_event(self):
        """Get a specific event."""
        events = self.api.get_events()
        assert events.cursor is not None
        self.generate_event()
        last_event = self.api.get_events(cursor=events.cursor).items[-1]
        assert last_event.id is not None
        event = self.api.get_event(id_=last_event.id)

        self.assertEqual(event.assigned_fields(), event.default_fields())
        self.assertEqual(event.type, dt.EventChangeType.CREATED)
        self.assertEqual(event.type_, dt.ItemType.ITEM_CHANGE)

        # TODO: What is the purpose?
        # for property_ in event.default_fields():
        #    # https://github.com/python/mypy/issues/7178
        #    self.assertEqual(event[property_], event[property_])  # type: ignore

    def test_get_events_by_cursor(self):
        """Get all events by specifying a cursor of 0."""
        previous_created_time = datetime(2000, 1, 1)
        previous_id = 0

        self.generate_event()
        events = self.api.get_events(cursor=0)

        for event in events.items:
            current_created_time = event.created_time
            assert current_created_time is not None
            assert event.id is not None
            self.assertGreater(current_created_time, previous_created_time)
            self.assertGreater(event.id, previous_id)
            previous_created_time = current_created_time
            previous_id = event.id
            self.assertEqual(event.assigned_fields(), event.default_fields())

    def test_get_events_empty(self):
        """
        If no cursor is given, the latest cursor is returned to retrieve future events.
        """
        events = self.api.get_events()
        self.assertEqual(events.items, [])
        self.assertFalse(events.has_more)

    def test_get_events_valid_properties(self):
        """Try to get specific properties of an event."""
        # TODO: Doesn't work with cursor=0
        property_combinations = self.get_combinations(dt.EventData.fields())
        for properties in property_combinations:
            events = self.api.get_events(fields=",".join(properties))
            for event in events.items:
                self.assertEqual(event.assigned_fields(), set(properties))


class Note(ClientBase):
    def test_add(self):
        """Add a note to an existing notebook."""
        parent_id = self.api.add_notebook()
        id_ = self.api.add_note()

        notes = self.api.get_notes().items
        self.assertEqual(len(notes), 1)
        self.assertEqual(notes[0].id, id_)
        self.assertEqual(notes[0].parent_id, parent_id)

    def test_add_attach_image(self):
        """Add a note with an image attached."""
        self.api.add_notebook()
        image_data = tools.encode_base64("test/grant_authorization_button.png")
        note_id = self.api.add_note(
            image_data_url=f"data:image/png;base64,{image_data}"
        )

        # Check that there is a resource.
        resources = self.api.get_resources().items
        self.assertEqual(len(resources), 1)
        resource_id = resources[0].id

        # Verify the resource is attached to the note.
        resources = self.api.get_resources(note_id=note_id).items
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0].id, resource_id)

    def test_add_no_notebook(self):
        """A note has to be added to an existing notebook."""
        with self.assertRaises(requests.exceptions.HTTPError) as context:
            self.api.add_note()
        self.assertEqual(context.exception.response.status_code, 500)
        self.assertIn(
            "Internal Server Error: Cannot find folder for note",
            context.exception.response.json()["error"],
        )

    def test_delete(self):
        """Add and then delete a note."""
        self.api.add_notebook()
        id_ = self.api.add_note()
        notes = self.api.get_notes()
        self.assertEqual(len(notes.items), 1)

        self.api.delete_note(id_=id_, permanent=1)
        self.assertEqual(self.api.get_notes().items, [])

    def test_get_note(self):
        """Get a specific note."""
        self.api.add_notebook()
        id_ = self.api.add_note()
        note = self.api.get_note(id_=id_)
        self.assertEqual(note.assigned_fields(), note.default_fields())
        self.assertEqual(note.type_, dt.ItemType.NOTE)

    def test_get_notes(self):
        """Get all notes."""
        self.api.add_notebook()
        self.api.add_note()
        notes = self.api.get_notes()
        self.assertEqual(len(notes.items), 1)
        self.assertFalse(notes.has_more)
        for note in notes.items:
            self.assertEqual(note.assigned_fields(), note.default_fields())

    def test_get_all_notes(self):
        """Get all notes, unpaginated."""
        self.api.add_notebook()
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        for _ in range(count):
            self.api.add_note()
        self.assertEqual(len(self.api.get_notes(limit=limit).items), min(limit, count))
        self.assertEqual(len(self.api.get_all_notes(limit=limit)), count)

    def test_get_notes_too_many_ids(self):
        """At maximum one parent ID can be used to obtain notes."""
        with self.assertRaises(ValueError):
            self.api.get_notes(notebook_id="1", tag_id="2")

    def test_get_notes_valid_properties(self):
        """Try to get specific properties of a note."""
        self.api.add_notebook()
        self.api.add_note()
        selected_fields = dt.NoteData.fields() - {
            # TODO: Some of the fields yield HTTP 500.
            "body_html",
            "base_url",
            "image_data_url",
            "crop_rect",
            # todo timestamps can be None
            "todo_due",
            "todo_completed",
            # we don't delete notes here
            "deleted_time",
        }
        property_combinations = self.get_combinations(selected_fields)
        for properties in property_combinations:
            notes = self.api.get_notes(fields=",".join(properties))
            for note in notes.items:
                self.assertEqual(note.assigned_fields(), set(properties))

    def test_move(self):
        """Move a note from one notebook to another."""
        original_title = "test title"
        original_body = "test body"

        notebook_1_id = self.api.add_notebook()
        notebook_2_id = self.api.add_notebook()
        id_ = self.api.add_note(
            parent_id=notebook_1_id, title=original_title, body=original_body
        )

        note = self.api.get_note(id_=id_)
        self.assertEqual(note.parent_id, notebook_1_id)

        self.api.modify_note(id_=id_, parent_id=notebook_2_id)
        note = self.api.get_note(id_=id_, fields="body,parent_id,title")
        self.assertEqual(note.parent_id, notebook_2_id)

        # Ensure the original properties aren't modified.
        self.assertEqual(note.body, original_body)
        self.assertEqual(note.title, original_title)


class Notebook(ClientBase):
    def test_add(self):
        """Add a notebook."""
        id_ = self.api.add_notebook()

        notebooks = self.api.get_notebooks().items
        self.assertEqual(len(notebooks), 1)
        self.assertEqual(notebooks[0].id, id_)

    def test_get_notebook(self):
        """Get a specific notebook."""
        id_ = self.api.add_notebook()
        notebook = self.api.get_notebook(id_=id_)
        self.assertEqual(notebook.assigned_fields(), notebook.default_fields())
        self.assertEqual(notebook.type_, dt.ItemType.FOLDER)

    def test_get_notebooks(self):
        """Get all notebooks."""
        self.api.add_notebook()
        notebooks = self.api.get_notebooks()
        self.assertEqual(len(notebooks.items), 1)
        self.assertFalse(notebooks.has_more)
        for notebook in notebooks.items:
            self.assertEqual(notebook.assigned_fields(), notebook.default_fields())

    def test_get_all_notebooks(self):
        """Get all notebooks, unpaginated."""
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        for _ in range(count):
            self.api.add_notebook()
        self.assertEqual(
            len(self.api.get_notebooks(limit=limit).items), min(limit, count)
        )
        self.assertEqual(len(self.api.get_all_notebooks(limit=limit)), count)

    def test_get_notebooks_invalid_property(self):
        """Try to get an non existent property of notebooks."""
        self.api.add_notebook()
        invalid_property = "invalid_property"
        with self.assertRaises(requests.exceptions.HTTPError) as context:
            self.api.get_notebooks(fields=invalid_property)
        self.assertEqual(context.exception.response.status_code, 500)
        self.assertIn(
            "Internal Server Error: Error: SQLITE_ERROR: no such column: "
            f"{invalid_property}",
            context.exception.response.json()["error"],
        )

    def test_get_notebooks_valid_properties(self):
        """Try to get specific properties of a notebook."""
        self.api.add_notebook()
        selected_fields = dt.NotebookData.fields() - {
            # we don't delete notebooks here
            "deleted_time",
        }
        property_combinations = self.get_combinations(selected_fields)
        for properties in property_combinations:
            notebooks = self.api.get_notebooks(fields=",".join(properties))
            for notebook in notebooks.items:
                self.assertEqual(notebook.assigned_fields(), set(properties))

    def test_move(self):
        """Move a root notebok to another notebook. It's now a subnotebook."""
        first_id = self.api.add_notebook()
        second_id = self.api.add_notebook()
        self.assertEqual(self.api.get_notebook(id_=first_id).parent_id, "")
        self.assertEqual(self.api.get_notebook(id_=second_id).parent_id, "")

        self.api.modify_notebook(id_=first_id, parent_id=second_id)
        self.assertEqual(self.api.get_notebook(id_=first_id).parent_id, second_id)
        self.assertEqual(self.api.get_notebook(id_=second_id).parent_id, "")


class Ping(ClientBase):
    def test_ping(self):
        """Ping should return the test string."""
        ping = self.api.ping()
        self.assertEqual(ping.text, "JoplinClipperServer")

    def test_ping_wrong_method(self):
        """Pinging the wrong method should return an error code."""
        for method in ("delete", "post", "put"):
            with self.subTest(method=method):
                with self.assertRaises(requests.exceptions.HTTPError) as context:
                    getattr(self.api, method)("/ping")
                self.assertEqual(context.exception.response.status_code, 405)


class Resource(ClientBase):
    @common.with_resource
    def test_add(self, filename):
        """Add a resource."""
        id_ = self.api.add_resource(filename=filename)

        resources = self.api.get_resources().items
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0].id, id_)

    @common.with_resource
    def test_add_to_note(self, filename):
        """Add a resource to an existing note."""
        self.api.add_notebook()
        for file_ in ["test/grant_authorization_button.png", filename]:
            with self.subTest(file_=file_):
                note_id = self.api.add_note()
                resource_id = self.api.add_resource(filename=file_)
                self.api.add_resource_to_note(resource_id=resource_id, note_id=note_id)

                # Verify the resource is attached to the note.
                resources = self.api.get_resources(
                    note_id=note_id, fields="id,mime"
                ).items
                self.assertEqual(len(resources), 1)
                self.assertEqual(resources[0].id, resource_id)

                # Verify the markdown is correct (prefix "!" for images).
                note = self.api.get_note(id_=note_id, fields="body")
                # TODO: Use "assertIsNotNone()" when
                # https://github.com/python/mypy/issues/5528 is resolved.
                assert resources[0].mime is not None
                image_prefix = "!" if resources[0].mime.startswith("image/") else ""
                self.assertEqual(
                    f"\n{image_prefix}[{file_}](:/{resource_id})", note.body
                )

    @common.with_resource
    def test_delete(self, filename):
        """Add and then delete a resource."""
        id_ = self.api.add_resource(filename=filename)
        resources = self.api.get_resources()
        self.assertEqual(len(resources.items), 1)

        self.api.delete_resource(id_=id_)
        self.assertEqual(self.api.get_resources().items, [])
        self.assertEqual(os.listdir(f"{PROFILE}/resources"), [])

    @common.with_resource
    def test_get_resource(self, filename):
        """Get metadata about a specific resource."""
        id_ = self.api.add_resource(filename=filename)
        resource = self.api.get_resource(id_=id_)
        self.assertEqual(resource.assigned_fields(), resource.default_fields())
        self.assertEqual(resource.type_, dt.ItemType.RESOURCE)

    @common.with_resource
    def test_get_resource_file(self, filename):
        """Get a specific resource in binary format."""
        for file_ in ("test/grant_authorization_button.png", filename):
            id_ = self.api.add_resource(filename=file_)
            resource = self.api.get_resource_file(id_=id_)
            with open(file_, "rb") as resource_file:
                self.assertEqual(resource, resource_file.read())

    @common.with_resource
    def test_get_resources(self, filename):
        """Get all resources."""
        self.api.add_resource(filename=filename)
        resources = self.api.get_resources()
        self.assertEqual(len(resources.items), 1)
        self.assertFalse(resources.has_more)
        for resource in resources.items:
            self.assertEqual(resource.assigned_fields(), resource.default_fields())

    @common.with_resource
    def test_get_all_resources(self, filename):
        """Get all resources, unpaginated."""
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        for _ in range(count):
            self.api.add_resource(filename=filename)
        self.assertEqual(
            len(self.api.get_resources(limit=limit).items), min(limit, count)
        )
        self.assertEqual(len(self.api.get_all_resources(limit=limit)), count)

    @common.with_resource
    def test_get_resources_valid_properties(self, filename):
        """Try to get specific properties of a resource."""
        self.api.add_resource(filename=filename)
        property_combinations = self.get_combinations(dt.ResourceData.fields())
        for properties in property_combinations:
            resources = self.api.get_resources(fields=",".join(properties))
            for resource in resources.items:
                self.assertEqual(resource.assigned_fields(), set(properties))

    @common.with_resource
    def test_modify_title(self, filename):
        """Modify a resource title."""
        id_ = self.api.add_resource(filename=filename)

        new_title = self.get_random_string()
        self.api.modify_resource(id_=id_, title=new_title)
        self.assertEqual(self.api.get_resource(id_=id_).title, new_title)

    @common.with_resource
    def test_check_derived_properties(self, filename):
        """Check the derived properties. I. e. mime type, extension and size."""
        for file_ in ["test/grant_authorization_button.png", filename]:
            id_ = self.api.add_resource(filename=file_)
            resource = self.api.get_resource(id_=id_, fields="mime,file_extension,size")
            mime_type, _ = mimetypes.guess_type(file_)
            self.assertEqual(
                resource.mime,
                mime_type if mime_type is not None else "application/octet-stream",
            )
            self.assertEqual(resource.file_extension, os.path.splitext(file_)[1][1:])
            self.assertEqual(resource.size, os.path.getsize(file_))

    @common.with_resource
    def test_check_property_title(self, filename):
        """Check the title of a resource."""
        title = self.get_random_string()
        id_ = self.api.add_resource(filename=filename, title=title)
        resource = self.api.get_resource(id_=id_)
        self.assertEqual(resource.title, title)


class Revision(ClientBase):
    def test_add(self):
        """Add a revision."""
        self.api.add_notebook()  # notebook for the note
        note_id = self.api.add_note()  # note that gets a new revision
        id_ = self.api.add_revision(note_id, dt.ItemType.NOTE)

        revisions = self.api.get_revisions().items
        self.assertEqual(len(revisions), 1)
        self.assertEqual(revisions[0].id, id_)

    def test_get_revision(self):
        """Get a specific revision."""
        self.api.add_notebook()  # notebook for the note
        note_id = self.api.add_note()  # note that gets a new revision
        id_ = self.api.add_revision(note_id, dt.ItemType.NOTE)
        revision = self.api.get_revision(id_=id_)
        self.assertEqual(revision.assigned_fields(), revision.default_fields())
        self.assertEqual(revision.type_, dt.ItemType.REVISION)

    def test_get_revisions(self):
        """Get all revisions."""
        self.api.add_notebook()  # notebook for the note
        note_id = self.api.add_note()  # note that gets a new revision
        self.api.add_revision(note_id, dt.ItemType.NOTE)
        revisions = self.api.get_revisions()
        self.assertEqual(len(revisions.items), 1)
        self.assertFalse(revisions.has_more)
        for revision in revisions.items:
            self.assertEqual(revision.assigned_fields(), revision.default_fields())

    def test_get_all_revisions(self):
        """Get all revisions, unpaginated."""
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        self.api.add_notebook()  # notebook for the note
        note_id = self.api.add_note()  # note that gets a new revision
        for _ in range(count):
            self.api.add_revision(note_id, dt.ItemType.NOTE)
        self.assertEqual(
            len(self.api.get_revisions(limit=limit).items), min(limit, count)
        )
        self.assertEqual(len(self.api.get_all_revisions(limit=limit)), count)

    def test_get_revisions_valid_properties(self):
        """Try to get specific properties of a revision."""
        self.api.add_notebook()  # notebook for the note
        note_id = self.api.add_note()  # note that gets a new revision
        self.api.add_revision(note_id, dt.ItemType.NOTE)
        property_combinations = self.get_combinations(dt.RevisionData.fields())
        for properties in property_combinations:
            revisions = self.api.get_revisions(fields=",".join(properties))
            for revision in revisions.items:
                self.assertEqual(revision.assigned_fields(), set(properties))


# TODO: Add more tests for the search parameter.
class Search(ClientBase):
    def test_empty(self):
        """Search should succeed, even if there is no result item."""
        self.assertEqual(self.api.search(query="*").items, [])

    def test_notes(self):
        """
        Wildcard search for all notes is disabled, because of performance reasons.
        See: https://github.com/laurent22/joplin/issues/5546
        """
        self.api.add_notebook()
        self.api.add_note()
        self.assertEqual(self.api.search(query="*").items, [])
        self.assertEqual(self.api.search(query="*", type="note").items, [])

    def test_notebooks(self):
        """Search by notebooks and search endpoint should yield same results."""
        self.api.add_notebook()
        self.assertEqual(
            self.api.search(query="*", type="folder"),
            self.api.get_notebooks(),
        )

    @common.with_resource
    def test_resources(self, filename):
        """Search by resources and search endpoint should yield same results."""
        self.api.add_resource(filename=filename)
        self.assertEqual(
            self.api.search(query="*", type="resource"),
            self.api.get_resources(),
        )

    def test_tags(self):
        """Search by tags and search endpoint should yield same results."""
        self.api.add_tag()
        self.assertEqual(
            self.api.search(query="*", type="tag"),
            self.api.get_tags(),
        )

    def test_master_key(self):
        """There is no master key configured."""
        self.assertEqual(
            self.api.search(query="*", type="master_key").items,
            [],
        )

    def test_non_searchable(self):
        """Check if the non searchable types throw an error."""
        # fmt: off
        for type_ in ("setting", "note_tag", "search", "alarm", "item_change",
                      "note_resource", "resource_local_state", "revision", "migration",
                      "smart_filter", "command"):
            with self.assertRaises(requests.exceptions.HTTPError) as context:
                self.api.search(query="*", type=type_)
            self.assertEqual(context.exception.response.status_code, 500)
        # fmt: on

    def test_pagination(self):
        """If there are more than 10 items, the results will be paginated."""
        limit = 10  # maximum 100

        for _ in range(limit + 1):
            self.api.add_notebook()

        query: dt.JoplinKwargs = {"query": "*", "type": "folder", "limit": limit}
        search_result = self.api.search(**query)
        self.assertEqual(len(search_result.items), limit)
        self.assertTrue(search_result.has_more)

        query["page"] = 2
        search_result = self.api.search(**query)
        self.assertEqual(len(search_result.items), 1)
        self.assertFalse(search_result.has_more)

    def test_search_query_special_chars(self):
        """
        Search should succeed even with special characters.
        See: https://github.com/marph91/joppy/issues/5
        """
        queries = (
            (
                "https://books.google.com.br/books?id=vaZFBgAAQBAJ&pg=PA83&dq=lakatos+"
                "copernicus&hl=pt-BR&sa=X&ved=0ahUKEwjewoWZ6q7hAhUNJ7kGHdy5CZUQ6AEIUDA"
                r"F#v=onepage&q=lakatos%20copernicus&f=false"
            ),
            (
                "https://www.facebook.com/photo.php?fbid=621934757910715&set=a.1437947"
                "52391387&type=3&app=fbl"
            ),
            r"foo# ?bar%+!baz",
            # Leading slashes in notebooks are stripped by default:
            # https://github.com/laurent22/joplin/issues/6213
            self.get_random_string().lstrip("/\\"),
        )

        for query in queries:
            self.api.add_notebook(title=query)
            result = self.api.search(query=query, type="folder")
            self.assertEqual(len(result.items), 1)
            self.assertEqual(result.items[0].title, query)

    def test_search_all(self):
        """Search notebooks and return all results, unpaginated."""
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        # Leading slashes in notebooks are stripped by default:
        # https://github.com/laurent22/joplin/issues/6213
        title = self.get_random_string().lstrip("/\\")
        query: dt.JoplinKwargs = {"query": title, "type": "folder", "limit": limit}
        for _ in range(count):
            self.api.add_notebook(title=title)
        self.assertEqual(len(self.api.search(**query).items), min(limit, count))
        self.assertEqual(len(self.api.search_all(**query)), count)


class Tag(ClientBase):
    def test_add_no_note(self):
        """Tags can be added even without notes."""
        id_ = self.api.add_tag()

        tags = self.api.get_tags().items
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].id, id_)

    def test_add_to_note(self):
        """Add a tag to an existing note."""
        self.api.add_notebook()
        note_id = self.api.add_note()
        tag_id = self.api.add_tag()
        self.api.add_tag_to_note(tag_id=tag_id, note_id=note_id)

        notes = self.api.get_notes(tag_id=tag_id).items
        self.assertEqual(len(notes), 1)
        self.assertEqual(notes[0].id, note_id)
        tags = self.api.get_all_tags()
        self.assertEqual(len(tags), 1)

    def test_add_with_parent(self):
        """Add a tag as child for an existing note."""
        self.api.add_notebook()
        parent_id = self.api.add_note()
        id_ = self.api.add_tag(parent_id=parent_id)

        tags = self.api.get_tags().items
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].id, id_)
        self.assertEqual(tags[0].parent_id, parent_id)

    def test_add_duplicated_name(self):
        """Tag names have to be unique."""
        # Note: Tags are always lower case:
        # https://discourse.joplinapp.org/t/tags-lower-case-only/4220
        # Note: Whitespace chars are substituted.
        tag_name = self.get_random_string(exclude=string.whitespace)

        self.api.add_tag(title=tag_name)
        with self.assertRaises(requests.exceptions.HTTPError) as context:
            self.api.add_tag(title=tag_name)
        self.assertEqual(context.exception.response.status_code, 500)
        self.assertIn(
            f'Internal Server Error: The tag "{tag_name.lower()}" already exists',
            context.exception.response.json()["error"],
        )

    def test_add_duplicated_id(self):
        """Tag IDs have to be unique."""
        tag_id = self.get_random_id()

        self.api.add_tag(id_=tag_id)
        with self.assertRaises(requests.exceptions.HTTPError) as context:
            self.api.add_tag(id_=tag_id)
        self.assertEqual(context.exception.response.status_code, 500)
        self.assertIn(
            "Internal Server Error: Error: SQLITE_CONSTRAINT: UNIQUE constraint failed",
            context.exception.response.json()["error"],
        )

    def test_get_tag(self):
        """Get a specific tag."""
        id_ = self.api.add_tag()
        tag = self.api.get_tag(id_=id_)
        self.assertEqual(tag.assigned_fields(), tag.default_fields())
        self.assertEqual(tag.type_, dt.ItemType.TAG)

    def test_get_tags(self):
        """Get all tags."""
        self.api.add_tag()
        tags = self.api.get_tags()
        self.assertEqual(len(tags.items), 1)
        self.assertFalse(tags.has_more)
        for tag in tags.items:
            self.assertEqual(tag.assigned_fields(), tag.default_fields())

    def test_get_all_tags(self):
        """Get all tags, unpaginated."""
        # Small limit and count to create/remove as less as possible items.
        count, limit = random.randint(1, 10), random.randint(1, 10)
        for _ in range(count):
            self.api.add_tag()
        self.assertEqual(len(self.api.get_tags(limit=limit).items), min(limit, count))
        self.assertEqual(len(self.api.get_all_tags(limit=limit)), count)

    def test_get_tags_valid_properties(self):
        """Try to get specific properties of a tag."""
        self.api.add_tag()
        property_combinations = self.get_combinations(dt.TagData.fields())
        for properties in property_combinations:
            tags = self.api.get_tags(fields=",".join(properties))
            for tag in tags.items:
                self.assertEqual(tag.assigned_fields(), set(properties))


class Fuzz(ClientBase):
    def test_random_path(self):
        """API should not crash, even with invalid paths."""
        for _ in range(1000 if common.SLOW_TESTS else 10):
            path = "/" + self.get_random_string(length=random.randint(0, 300))
            method = random.choice(("delete", "get", "post", "put"))
            try:
                self.api._request(method, path)
            except (
                requests.exceptions.HTTPError,
                urllib3.exceptions.LocationParseError,
            ):
                pass
        self.api.ping()


class Helper(ClientBase):
    """Tests for the helper functions."""

    def test_random_id(self):
        """Random IDs should always be valid."""
        for _ in range(100):
            self.assertTrue(dt.is_id_valid(self.get_random_id()))

    def test_is_id_valid(self):
        """Trivial test for the is_id_valid() function."""
        self.assertTrue(dt.is_id_valid("0" * 32))

        # ID has to be 32 chars.
        self.assertFalse(dt.is_id_valid("0" * 31))

        # ID has to be contain only hex chars.
        self.assertFalse(dt.is_id_valid("h" + "0" * 31))


class Miscellaneous(ClientBase):
    @common.with_resource
    def test_same_id_different_type(self, filename):
        """Same IDs can be used if the types are different."""
        id_ = self.get_random_id()

        self.api.add_notebook(id_=id_)
        self.api.add_note(id_=id_)
        self.api.add_resource(id_=id_, filename=filename)
        self.api.add_tag(id_=id_)


class Regression(ClientBase):
    @unittest.skip("Enable when the bug is fixed.")
    @unittest.skipIf(not common.SLOW_TESTS, "Generating the long string is slow.")
    def test_long_body(self):
        """
        https://github.com/laurent22/joplin/issues/5543
        Response HTTP 104: https://stackoverflow.com/a/52826181
        """
        # Use only one random character, since it's very slow already.
        body = self.get_random_string(1) * 10**9
        self.api.add_notebook()
        note_id = self.api.add_note(body=body)
        self.assertEqual(self.api.get_note(id_=note_id).title, body)

    def test_note_tag_fields(self):
        """https://github.com/laurent22/joplin/issues/4407"""
        self.api.add_notebook()
        note_id = self.api.add_note()
        tag_id = self.api.add_tag()
        self.api.add_tag_to_note(tag_id=tag_id, note_id=note_id)

        notes = self.api.get_notes(tag_id=tag_id, fields="id")
        self.assertEqual(notes.items[0].assigned_fields(), {"id"})

    @unittest.skip("Not yet implemented")
    def test_set_location(self):
        """https://github.com/laurent22/joplin/issues/3884"""

    @unittest.skip("Not yet implemented")
    def test_add_todo(self):
        """https://github.com/laurent22/joplin/issues/1687"""


class ReadmeExamples(ClientBase):
    """Check the readme examples for functionality."""

    readme_content: str = ""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        with open("README.md", encoding="utf-8") as infile:
            readme_content = infile.read()

        # Replace the token to make the code functional.
        cls.readme_content = readme_content.replace("YOUR_TOKEN", f'"{API_TOKEN}"')

    def get_example_code(self, example_name: str) -> str:
        """
        Get the code of a readme example by its name.
        The example can be identified by a custom info string in markdown.
        See: https://spec.commonmark.org/0.30/#example-143
        """

        matches = re.search(
            f"```python name={example_name}\n(.*?)```",
            self.readme_content,
            flags=re.DOTALL,
        )
        # TODO: Use "assertIsNotNone()" when
        # https://github.com/python/mypy/issues/5528 is resolved.
        assert matches is not None
        self.assertEqual(len(matches.groups()), 1)
        return matches.groups()[0]

    def test_get_all_notes(self):
        note_count = 3
        self.api.add_notebook()
        for _ in range(note_count):
            self.api.add_note()

        # Execute the example code. The local variables are stored in "locals_dict".
        code = self.get_example_code("get_all_notes")
        locals_dict: Mapping[str, Any] = {}
        exec(code, None, locals_dict)

        self.assertEqual(len(locals_dict["notes"]), note_count)

    def test_add_tag_to_note(self):
        code = self.get_example_code("add_tag_to_note")
        exec(code)

        tags = self.api.get_all_tags()
        self.assertEqual(len(tags), 1)

        assert tags[0].id is not None
        notes = self.api.get_all_notes(tag_id=tags[0].id)
        self.assertEqual(len(notes), 1)

    def test_add_resource_to_note(self):
        code = self.get_example_code("add_resource_to_note")
        code = code.replace("path/to/image.png", "test/grant_authorization_button.png")
        exec(code)

        notes = self.api.get_all_notes()
        self.assertEqual(len(notes), 2)

        resources = self.api.get_all_resources()
        self.assertEqual(len(resources), 2)

        # Each note should reference to exactly one resource.
        for note in notes:
            assert note.id is not None
            resources = self.api.get_all_resources(note_id=note.id)
            self.assertEqual(len(resources), 1)

    def test_remove_tags(self):
        self.api.add_tag(title="Title")
        self.api.add_tag(title="! Another Title")
        self.api.add_tag(title="!_third_title")

        code = self.get_example_code("remove_tags")
        exec(code)

        # All tags starting with "!" should be removed.
        tags = self.api.get_all_tags()
        self.assertEqual(len(tags), 1)
        self.assertEqual(tags[0].title, "title")  # tags are always lower case

    def test_remove_spaces_from_tags(self):
        self.api.add_tag(title="tag with spaces")
        self.api.add_tag(title="another tag with spaces")

        code = self.get_example_code("remove_spaces_from_tags")
        exec(code)

        all_tags = self.api.get_all_tags()
        self.assertEqual(len(all_tags), 2)
        for tag in all_tags:
            assert tag.title is not None
            self.assertNotIn(" ", tag.title)

    @common.with_resource
    def test_remove_orphaned_resources(self, filename):
        self.api.add_notebook()
        for i in range(2):
            note_id = self.api.add_note()
            resource_id = self.api.add_resource(
                filename=filename, title=f"resource {i}"
            )
            self.api.add_resource_to_note(resource_id=resource_id, note_id=note_id)

        # Delete the second note, which creates an orphaned resource.
        self.api.delete_note(note_id, permanent=1)
        self.assertEqual(len(self.api.get_all_resources()), 2)

        code = self.get_example_code("remove_orphaned_resources")
        exec(code)

        # The resource without reference should be deleted.
        resources = self.api.get_all_resources()
        self.assertEqual(len(resources), 1)
        self.assertEqual(resources[0].title, "resource 0")
